# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pb/headers.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10pb/headers.proto\x12\x18\x63om.arpedon.measurements\"\xf4\x01\n\x11MeasurementHeader\x12\x13\n\x0bsample_rate\x18\x01 \x02(\x05\x12K\n\x08\x63hannels\x18\x02 \x03(\x0b\x32\x39.com.arpedon.measurements.MeasurementHeader.ChannelHeader\x1a}\n\rChannelHeader\x12\x11\n\treference\x18\x01 \x02(\x02\x12\x19\n\x0bsensitivity\x18\x02 \x01(\x02:\x04\x31\x30\x30\x30\x12\x17\n\x0c\x64\x62_reference\x18\x03 \x01(\x02:\x01\x30\x12\x12\n\x07pregain\x18\x04 \x01(\x02:\x01\x30\x12\x11\n\x06offset\x18\x05 \x01(\x02:\x01\x30\x42\x1a\n\x18\x63om.arpedon.measurements')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pb.headers_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.arpedon.measurements'
  _globals['_MEASUREMENTHEADER']._serialized_start=47
  _globals['_MEASUREMENTHEADER']._serialized_end=291
  _globals['_MEASUREMENTHEADER_CHANNELHEADER']._serialized_start=166
  _globals['_MEASUREMENTHEADER_CHANNELHEADER']._serialized_end=291
# @@protoc_insertion_point(module_scope)
